from cStringIO import StringIO

import datetime
import gzip
import inspect
import logging
import os
import sys
import traceback
import urllib
import urllib2

from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed, ImproperlyConfigured
from django.core.urlresolvers import resolve

from djexceptional.utils import memoize, json_dumps, meta_to_http


__version__ = '0.1.4'

EXCEPTIONAL_PROTOCOL_VERSION = 6
EXCEPTIONAL_API_ENDPOINT = getattr(settings, 'EXCEPTIONAL_API_ENDPOINT',
                                   "http://api.getexceptional.com/api/errors")

LOG = logging.getLogger('djexceptional')


class ExceptionalMiddleware(object):

    """
    Middleware to interface with the Exceptional service.

    Requires very little intervention on behalf of the user; you just need to
    add `EXCEPTIONAL_API_KEY` to your Django settings. You can also optionally
    set `EXCEPTIONAL_API_ENDPOINT` to change the API endpoint which will be
    used; the default is `'http://api.getexceptional.com/api/errors'`.
    """

    def __init__(self):
        if settings.DEBUG:
            raise MiddlewareNotUsed

        try:
            self.api_key = settings.EXCEPTIONAL_API_KEY
        except AttributeError:
            raise ImproperlyConfigured("You need to add an EXCEPTIONAL_API_KEY setting.")

        self.api_endpoint = EXCEPTIONAL_API_ENDPOINT + "?" + urllib.urlencode({
            "api_key": self.api_key,
            "protocol_version": EXCEPTIONAL_PROTOCOL_VERSION
            })

    def process_exception(self, request, exc):
        info = {}
        info.update(self.environment_info())
        info.update(self.request_info(request))
        info.update(self.exception_info(exc, sys.exc_info()[2]))

        payload = self.compress(json_dumps(info))
        req = urllib2.Request(self.api_endpoint, data=payload)
        req.headers['Content-Encoding'] = 'gzip'
        req.headers['Content-Type'] = 'application/json'

        try:
            conn = urllib2.urlopen(req)
            try:
                conn.read()
            finally:
                conn.close()
        except Exception, exc:
            LOG.exception("Error communicating with the Exceptional service: %r", exc)

    @staticmethod
    def compress(bytes):
        """Compress a bytestring using gzip."""

        stream = StringIO()
        # Use `compresslevel=1`; it's the least compressive but it's fast.
        gzstream = gzip.GzipFile(fileobj=stream, compresslevel=1, mode='wb')
        try:
            try:
                gzstream.write(bytes)
            finally:
                gzstream.close()
            return stream.getvalue()
        finally:
            stream.close()

    @memoize
    def environment_info(self):

        """
        Return a dictionary representing the server environment.

        The idea is that the result of this function will rarely (if ever)
        change for a given app instance. Ergo, the result can be cached between
        requests.
        """

        return {
                "application_environment": {
                    "framework": "django",
                    "env": dict(os.environ),
                    "language": "python",
                    "language_version": sys.version.replace('\n', ''),
                    "application_root_directory": self.project_root()
                    },
                "client": {
                    "name": "django-exceptional",
                    "version": __version__,
                    "protocol_version": EXCEPTIONAL_PROTOCOL_VERSION
                    }
                }

    def request_info(self, request):

        """
        Return a dictionary of information for a given request.

        This will be run once for every request.
        """

        # We have to re-resolve the request path here, because the information
        # is not stored on the request.
        view, args, kwargs = resolve(request.path)
        for i, arg in enumerate(args):
            kwargs[i] = arg

        view_name = self.get_view_name(view)

        parameters = {}
        parameters.update(kwargs)
        parameters.update(request.POST.items())
        parameters = self.filter_params(parameters)

        return {
                "request": {
                    "session": dict(request.session),
                    "remote_ip": request.META["REMOTE_ADDR"],
                    "parameters": parameters,
                    "controller": view_name[0],
                    "action": view_name[1],
                    "url": request.build_absolute_uri(),
                    "request_method": request.method,
                    "headers": meta_to_http(request.META)
                    }
                }

    def exception_info(self, exception, tb, timestamp=None):
        backtrace = []
        for tb_part in traceback.format_tb(tb):
            backtrace.extend(tb_part.rstrip().splitlines())

        if timestamp is None:
            timestamp = datetime.datetime.utcnow()

        return {
                "exception": {
                    # Naively assume all times are in UTC.
                    "occurred_at": timestamp.isoformat() + 'Z',
                    "message": str(exception),
                    "backtrace": backtrace,
                    "exception_class": self.exception_class(exception)
                    }
                }

    def exception_class(self, exception):
        """Return a name representing the class of an exception."""

        cls = type(exception)
        if cls.__module__ == 'exceptions':  # Built-in exception.
            return cls.__name__
        return "%s.%s" % (cls.__module__, cls.__name__)

    @memoize
    def project_root(self):

        """
        Return the root of the current Django project on the filesystem.

        Looks for `settings.PROJECT_ROOT`; failing that, uses the directory
        containing the current settings file.
        """

        if hasattr(settings, 'PROJECT_ROOT'):
            return settings.PROJECT_ROOT

        settings_file = sys.modules[settings.SETTINGS_MODULE].__file__
        if settings_file.endswith(".pyc"):
            return settings_file[:-1]
        return settings_file

    @staticmethod
    def get_view_name(view):
        """Resolve a Django view object into a controller/action name pair."""

        if inspect.isfunction(view):
            # function_module, function_name
            return view.__module__, view.__name__
        elif inspect.ismethod(view):
            # class_module.ClassName, method_name
            return (view.im_class.__module__ + '.' + view.im_class.__name__,
                    view.__name__)
        # class_module, ClassName
        return view.__class__.__module__, view.__class__.__name__

    @staticmethod
    def filter_params(params):
        """Filter sensitive information out of parameter dictionaries."""

        for key in params.keys():
            if "password" in key:
                del params[key]
        return params



from django.test import TestCase

from djexceptional.utils import memoize


class MemoizeTest(TestCase):

    def test_no_args(self):
        """Test @memoize on functions without any arguments."""

        counter = []
        def increment_counter():
            counter.append(None)
            return len(counter)

        self.assertEqual(len(counter), 0)
        self.assertEqual(increment_counter(), 1)
        self.assertEqual(len(counter), 1)
        self.assertEqual(increment_counter(), 2)
        self.assertEqual(len(counter), 2)

        increment_counter = memoize(increment_counter)

        self.assertEqual(increment_counter(), 3)
        self.assertEqual(len(counter), 3)
        self.assertEqual(increment_counter(), 3)
        self.assertEqual(len(counter), 3)

    def test_args(self):
        """Test @memoize on functions with arguments."""

        counter = []
        def sum_squared(x, y):
            counter.append(None)
            return x ** 2 + y ** 2

        self.assertEqual(len(counter), 0)
        self.assertEqual(sum_squared(3, 4), 25)
        self.assertEqual(len(counter), 1)
        self.assertEqual(sum_squared(4, 5), 41)
        self.assertEqual(len(counter), 2)

        sum_squared = memoize(sum_squared)

        self.assertEqual(sum_squared(3, 4), 25)
        self.assertEqual(len(counter), 3)
        self.assertEqual(sum_squared(3, 4), 25)
        self.assertEqual(sum_squared(3, 4), 25)
        self.assertEqual(sum_squared(3, 4), 25)
        self.assertEqual(len(counter), 3)

        self.assertEqual(sum_squared(4, 5), 41)
        self.assertEqual(len(counter), 4)
        self.assertEqual(sum_squared(4, 5), 41)
        self.assertEqual(sum_squared(4, 5), 41)
        self.assertEqual(sum_squared(4, 5), 41)
        self.assertEqual(len(counter), 4)

    def test_clear(self):
        """Test the `clear()` method added to wrapped functions."""

        counter = []
        def increment_counter():
            counter.append(None)
            return len(counter)
        increment_counter = memoize(increment_counter)

        self.assertEqual(len(counter), 0)
        self.assertEqual(increment_counter(), 1)
        self.assertEqual(len(counter), 1)
        self.assertEqual(increment_counter(), 1)
        self.assertEqual(increment_counter(), 1)
        self.assertEqual(increment_counter(), 1)
        self.assertEqual(len(counter), 1)

        increment_counter.clear()

        self.assertEqual(increment_counter(), 2)
        self.assertEqual(len(counter), 2)
        self.assertEqual(increment_counter(), 2)
        self.assertEqual(increment_counter(), 2)
        self.assertEqual(increment_counter(), 2)
        self.assertEqual(len(counter), 2)



import datetime
import decimal
import re

from django.utils import datetime_safe
from django.utils import simplejson


class ResilientJSONEncoder(simplejson.JSONEncoder):
    """A JSON encoder (with support for dates/times) that should never fail."""

    DATE_FORMAT = "%Y-%m-%d"
    TIME_FORMAT = "%H:%M:%S"

    def default(self, o):
        if isinstance(o, datetime.datetime):
            d = datetime_safe.new_datetime(o)
            return d.strftime("%s %s" % (self.DATE_FORMAT, self.TIME_FORMAT))
        elif isinstance(o, datetime.date):
            d = datetime_safe.new_date(o)
            return d.strftime(self.DATE_FORMAT)
        elif isinstance(o, datetime.time):
            return o.strftime(self.TIME_FORMAT)
        elif isinstance(o, decimal.Decimal):
            return str(o)
        else:
            try:
                return super(ResilientJSONEncoder, self).default(o)
            except Exception:
                return repr(o)


def json_dumps(obj):
    """Dump an object to a JSON string, using the resilient JSON encoder."""

    return simplejson.dumps(obj, cls=ResilientJSONEncoder)


def meta_to_http(meta):
    """Convert a request.META into a dictionary of HTTP headers."""

    headers = {}
    for key in meta:
        if key.startswith("HTTP_"):
            # A heuristic; HTTP_X_FORWARDED_FOR => X-Forwarded-For
            header = re.sub(r'^HTTP_', '', key)
            header = key.replace("_", " ").title().replace(" ", "-")
        elif key in ["CONTENT_LENGTH", "CONTENT_TYPE"]:
            header = key.replace("_", " ").title().replace(" ", "-")
        else:
            continue
        headers[header] = meta[key]
    return headers


def memoize(func):
    """A simple memoize decorator (with no support for keyword arguments)."""

    cache = {}
    def wrapper(*args):
        if args in cache:
            return cache[args]
        cache[args] = value = func(*args)
        return value

    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    if hasattr(func, '__module__'):
        wrapper.__module__ = func.__module__
    wrapper.clear = cache.clear

    return wrapper
